# FILE: setBackupFolder.tcl
#
# LAST UPDATE: 01/24/93 3:46:17 PM

# setBackupFolder = sets the backup folder based on location of file being
# edited.  If folder named "$backName" is found relative to the file's location
# or relative to any folder above, then that is used.  If no "$backName" folder
# is found, then turn off backupFolder.  Thus the following hypothetical volume
# has respective backup files:
#
# HD
# :Backup:
# ::Addresses.bak
# :Lists:
# ::Addresses
# :Misc:
# ::Letters:
# :::Project RFQ
# ::Backup:
# :::Project RFQ.bak
# ::Memos:
# :::Backup:
# ::::Organization.bak
# :::Organization
# Floppy:
# :Readme
# :Readme.bak
#
# To use, source this file.  It redefines activateHook & saveasHook,
# keeping the old definitions as original-activateHook & original-saveHook.
# A good place to source this is either from $HOME:AlphaBits.tcl (at the
# end) or $HOME:UserBits.tcl

# COPYRIGHT:
#
#	Copyright  1992,1993 by David C. Black All rights reserved.
#	Portions copyright  1990, 1991, 1992 Pete Keleher. All Rights Reserved.
#
#	Redistribution and use in source and binary forms are permitted
#	provided that the above copyright notice and this paragraph are
#	duplicated in all such forms and that any documentation,
#	advertising materials, and other materials related to such
#	distribution and use acknowledge that the software was developed
#	by David C. Black in conjunction with Pete Keleher.
#
#	THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
#	IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
################################################################################

# AUTHOR
#
#	David C. Black
#	Internet: black@mpd.tandem.com (preferred)
#	GEnie:    D.C.Black
#	USnail:   6217 John Chisum Lane, Austin, TX 78749
#
################################################################################

set backupDir ":BackUp"
proc setBackupFolder {} {
	global backupDir
	global backupFolder
	global backupFolderName
	set wins [winNames -f]
	if {[llength $wins] == 0} {
		return
	} else {
		set folder [lindex $wins 0]
	}
	if {[regexp {^Untitled [0-9]+$} $folder]} {
		return
	}
	if {[regexp {^[*].*[*]$} $folder]} {
		# Ignore "*tcl shell*" and the like
		return
	}
	while {[string first ":" $folder] >= 0} {
		set folder  [file dirname $folder]
		if ([file isdirectory "$folder$backupDir"]) {
			set backupFolderName "$folder$backupDir"
			set backupFolder 1
			# message $backupFolderName
			return
		}
	}
	append folder "$backupDir"
	if ([file isdirectory "$folder"]) {
		set backupFolderName "$folder"
		set backupFolder 1
	} else {
		set backupFolder 0
	}
	# message $backupFolderName
}

if {[info procs original-saveasHook] != "original-saveasHook"} {

rename saveasHook original-saveasHook

proc saveasHook {oldName newName} {
	setBackupFolder
	set result [eval original-saveasHook {$oldName} {$newName}]
	return $result
}
#endproc saveasHook

rename activateHook original-activateHook

proc activateHook name {
	setBackupFolder
	set result [eval original-activateHook {$name}]
	return $result
}
#endproc activateHook

}
#endif

setBackupFolder
